

import calendar
import datetime
import email
import logging
import mimetypes
import os
import stat
import subprocess

import tornado.web
import assetman


class AssetmanMixin(object):
    def __init__(self, *args, **kwargs):
        super(AssetmanMixin, self).__init__(*args, **kwargs)
        assert hasattr(self.application, 'assetman_template_helper')
        assert isinstance(self.application.assetman_template_helper, 
            assetman.tornadoutils.helpers.TemplateCommands) 

    def render_string(self, template_name, **kwargs):
        return super(AssetmanMixin, self).render_string(template_name, 
            assetman=self.application.assetman_template_helper, **kwargs)



class StaticFileHandler(tornado.web.RequestHandler):

    def initialize(self, root, expires=True):
        assert isinstance(expires, bool)
        self.root = root
        self.expires = expires

    def head(self, path):
        return self.get(path, include_body=False)

    def get(self, path, include_body=True):
        abs_path = os.path.normpath(os.path.join(self.root, path))

        if not os.path.isfile(abs_path) or not abs_path.startswith(self.root):
            raise tornado.web.HTTPError(404, 'File %s not found', path)

        stat_result = os.stat(abs_path)
        modified = datetime.datetime.utcfromtimestamp(stat_result[stat.ST_MTIME])
        self.set_header("Last-Modified", modified)
        self.set_expires_header()
        self.set_mime_type(abs_path)

        # Check the If-Modified-Since, and don't send the result if the
        # content has not been modified
        ims_value = self.request.headers.get("If-Modified-Since")
        if ims_value is not None:
            date_tuple = email.utils.parsedate(ims_value)
            if_since = datetime.datetime.utcfromtimestamp(calendar.timegm(date_tuple))
            if if_since >= modified:
                logging.debug('Not modified since %s', if_since)
                self.set_status(304)
                return

        if not include_body:
            return

        self.set_header("Content-Length", stat_result[stat.ST_SIZE])
        with open(abs_path, 'rb') as f:
            logging.debug('Response headers: %r', self._headers)
            self.write(f.read())

    def set_expires_header(self):
        if self.expires:
            expires_at = datetime.datetime.utcnow() + datetime.timedelta(days=365*10)
            max_age = 86400 * 365 * 10
            self.set_header("Expires", expires_at)
            self.set_header("Cache-Control", "public, max-age=%s" % max_age)
        else:
            # no-cache == revalidate before serving from cache; different from
            # no-store
            self.set_header("Cache-Control", "no-cache")

    def set_mime_type(self, url):
        mime_type, encoding = mimetypes.guess_type(url)
        if not mime_type and url.endswith('.otf'):
            mime_type = 'application/octet-stream'
        if not mime_type and url.endswith('.ttf'):
            mime_type = 'font/ttf'
        if not mime_type and url.endswith('.eot'):
            mime_type = 'application/vnd.ms-fontobject'
        if not mime_type and url.endswith('.woff'):
            mime_type = 'application/x-font-woff'
        if not mime_type and url.endswith('.json'):
            mime_type = 'application/json'
        if not mime_type and url.endswith('.svg'):
            mime_type = 'image/svg+xml'
        if not mime_type and url.endswith('.csv'):
            mime_type = 'text/csv'
        if mime_type:
            self.set_header("Content-Type", mime_type)
            if mime_type == 'text/csv':
                filename = os.path.basename(url)
                disposition = "attachment; filename=%s" % filename
                self.set_header("Content-Disposition", disposition)
        else:
            logging.warning('Unable to guess mime type for %r', url)


class CompilingStaticHandler(tornado.web.RequestHandler):

    content_type = None

    def initialize(self, input_root, output_root=None):
        """Input root is where we will look for source files and output root
        is where we will put compiled/intermediate files (if necessary).
        """
        assert self.content_type, 'content_type must be set on %s' % self
        assert os.path.isdir(input_root), 'Input root %r must be a directory' % input_root
        if output_root is not None:
            assert os.path.isdir(output_root), 'Output root %r must be a directory' % output_root
        self.input_root = input_root
        self.output_root = output_root

    def get(self, path):
        if not self.settings['assetman_settings']['enable_static_compilation']:
            logging.error('Static compiler handler not enabled')
            raise tornado.web.HTTPError(404)

        abs_path = os.path.normpath(os.path.join(self.input_root, path))
        if not os.path.isfile(abs_path) or not abs_path.startswith(self.input_root):
            raise tornado.web.HTTPError(404, 'File %s not found', path)

        logging.debug('Compiling %s', abs_path)
        output = self.do_compile(abs_path, path)
        self.set_header('Content-Type', self.content_type)
        self.set_header('Cache-Control', 'no-cache')
        self.write(output)

    def do_compile(self, abs_path, url_path):
        raise NotImplementedError

    def run_proc(self, cmd, stdin=None, env=None):
        """Runs the given cmd as a subprocess, where cmd is a list suitable
        for passing to subprocess.call. Returns a 3-tuple of

            (exit code, stdout, stderr)
        """
        assert isinstance(cmd, (list, tuple))
        popen_args = dict(stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if stdin is not None:
            popen_args['stdin'] = subprocess.PIPE
        if env is not None:
            new_env = os.environ.copy()
            new_env.update(env)
            popen_args['env'] = new_env
        proc = subprocess.Popen(cmd, **popen_args)
        out, err = proc.communicate(input=stdin)

        if proc.returncode == 0:
            return out
        logging.error('Error in %r %r', ' '.join(cmd), err)
        raise tornado.web.HTTPError(500)


class LessCompilerHandler(CompilingStaticHandler):

    content_type = 'text/css'

    def do_compile(self, abs_path, url_path):
        cmd = [self.settings['assetman_settings']['lessc_path'], abs_path]
        env = {
            'PATH': os.environ.get('PATH', '') 
        }
        return self.run_proc(cmd, env=env)


class SassCompilerHandler(CompilingStaticHandler):

    content_type = 'text/css'

    def do_compile(self, abs_path, url_path):
        # We need to change into our input_root directory, which is where the
        # special config.rb file lives, which tells Compass how to compile
        # things.
        old_cwd = os.getcwd()
        os.chdir(self.input_root)
        cmd = [
            self.settings['assetman_settings']['sass_compiler_path'],
            '--compass', '--trace', '-l', '--stop-on-error',
            abs_path,
        ]
        try:
            out = self.run_proc(cmd)
            return out
        finally:
            os.chdir(old_cwd)




from assetman.settings import Settings
from assetman.compilers import JSCompiler, LessCompiler, CSSCompiler, SassCompiler

# Map from template-side assetman manager calls to the corresponding
# compiler classes
compiler_classes = [JSCompiler, LessCompiler, CSSCompiler, SassCompiler]
compiler_map = dict((c.include_expr, c) for c in compiler_classes)


class TemplateParser(object):
    """ Base template parsing class, to serve as a drop in template wrapper for
    template agnosticism.
    """

    def __init__(self, template_path, settings=None, **kwargs):
        self.settings = settings or Settings()
        self.template_path = template_path
        self.load_template(template_path)

    def load_template(self, path):
        """
        This method must be defined to load template-language specific loading
        mechanism on init.
        """
        raise NotImplementedError

    def get_compilers(self):
        """
        Define this method to return all compilers required for the assetman blocks
        contained by this template.
        """
        raise NotImplementedError





import os
import logging
import functools
import hashlib

from assetman.tools import get_shard_from_list, _unicode
from assetman.manifest import Manifest

class AssetManager(object):
    """AssetManager attempts to provide easy-to-use asset management and
    compilation for Tornado (or other?) templates.

    On the template side, assuming this `assetman` module is available in the
    template context, use in Tornado should be as easy as:

        {% apply assetman.include_js %}
        js/utils.js
        js/lib.js
        js/main.js
        {% end %}

    (Variations can/should be created for other frameworks.)

    With this block in place, each individual JavaScript file will be included
    in the resulting document in development. In production, a single,
    compiled JavaScript will be included instead.
    """

    def __init__(self, rel_url_text, local=False, include_tag=True, src_path=None, settings=None, **attrs):
        """Creates an asset manager from `rel_url_text`, a string which should
        contain space (or newline) separated asset URLs.

        Optional params:

          * local - if True, URLs in rendered output will point at our own
            hosts instead of our CDN.

          * include_tag - if False, the only rendered output will be URLs,
            without full HTML elements (useful, e.g., for mobile app manifests)

          * src_path - an optional annotation for recording where this asset
            manager originated.

        Any extra kwargs will be interpreted as extra HTML params to include
        on the rendered element.
        """
        self.rel_urls = [_f for _f in _unicode(rel_url_text).split() if _f]
        self.local = local
        self.include_tag = include_tag
        self.src_path = src_path
        self.attrs = attrs
        self._manifest = None
        assert settings
        self.settings = settings
        logging.debug('%s URLs: %r', self.__class__.__name__, self.rel_urls)

    # Lazy-load the manifest attribute
    def get_manifest(self):
        if not self._manifest:
            self._manifest = Manifest(self.settings).load()
        return self._manifest

    def set_manifest(self, manifest):
        self._manifest = manifest

    manifest = property(get_manifest, set_manifest)

    def get_hash(self):
        """Gets the md5 hash for the URLs in this block of assets, which will
        be used to refer to the compiled assets in production.
        """
        return hashlib.md5('\n'.join(self.rel_urls).encode()).hexdigest()

    def get_ext(self):
        """Returns the file extension (without leading period) to use for the
        compiled version of these assets.
        """
        raise NotImplementedError

    def get_compiled_name(self):
        """Returns the filename for the compiled version of this asset bundle,
        which is composed of its version hash and a filed extension.
        """
        name_hash = self.get_hash()
        return self.manifest.blocks[name_hash]['versioned_path']

    def make_asset_url(self, rel_url):
        """Builds a full URL based the given relative URL."""
        if self.settings['enable_static_compilation']:
            prefix = self.settings['static_url_prefix']
        elif self.local:
            prefix = self.settings.get('local_cdn_url_prefix')
        else:
            prefix = get_shard_from_list(self.settings['cdn_url_prefix'], os.path.basename(rel_url))
        return prefix.rstrip('/') + '/' + rel_url.lstrip('/')

    def render_attrs(self):
        """Returns this asset block's attrs as an HTML string. Includes a
        leading space.
        """
        attrs = ' '.join('%s=%r' % (attr, _unicode(val))
                         for attr, val in self.attrs.items())
        return ' ' + attrs if attrs else ''

    def render_asset(self, url):
        """Renders an individual asset at the given URL. The given URL should
        be the full URL to of the asset.
        """
        if self.include_tag:
            return self.render_asset_element(url)
        else:
            return url

    def render_asset_element(self, url):
        raise NotImplementedError

    def render(self):
        """Renders these assets. If static compilation is enabled, each asset is
        rendered individually. In a production environment, this should be disabled and
        just the compiled asset should rendered.
        """
        try:
            if self.settings['enable_static_compilation']:
                urls = list(map(self.make_asset_url, self.rel_urls))
                return '\n'.join(map(self.render_asset, urls))
            else:
                compiled_name = self.get_compiled_name()
                url = self.make_asset_url(compiled_name)
                return self.render_asset(url)
        except:
            logging.error('failed getting assets for %s', self.rel_urls)
            raise

    @classmethod
    def include(cls, s=None, **kwargs):
        """A shortcut for creating and rendering an AssetManager, which needs
        to support two styles of invocation in templates:

            {% apply assetman.include_js %}path/to/lib.js{% end %}

        and

            {% apply assetman.include_js(local=True) %}path/to/lib.js{% end %}

        Ie, a "bare" application without any args and one with args. In the
        latter case, we need to return a partial function that can be applied
        to a string.
        """
        if s is None:
            return functools.partial(cls.include, **kwargs)
        return cls(s, **kwargs).render()

    def static_url(self, url_path, local=None):
        """A shortcut for ensuring that the given URL is versioned in production.
        """
        if local is not None:
            self.local = local
        
        if self.settings['enable_static_compilation']:
            return self.make_asset_url(url_path)
        else:
            assert url_path in self.manifest.assets, url_path
            versioned_path = self.manifest.assets[url_path]['versioned_path']
            return self.make_asset_url(versioned_path)

    def __str__(self):
        return '<%s src:%s assets:%s>' % (
            self.__class__.__name__,
            self.src_path or '<template>',
            len(self.rel_urls))


class JSManager(AssetManager):
    def get_ext(self):
        return 'js'
    def render_asset_element(self, url):
        return '<script src="%s" type="text/javascript"%s></script>' % (
            url, self.render_attrs())


class CSSManager(AssetManager):
    def get_ext(self):
        return 'css'
    def render_asset_element(self, url):
        return '<link href="%s" rel="stylesheet" type="text/css"%s>' % (
            url, self.render_attrs())


class LessManager(CSSManager):
    # No customization needed because the custom static asset compiler handler
    # will take the *.less URL and return the CSS output.
    pass


class SassManager(CSSManager):
    # No customization needed because the custom static asset compiler handler
    # will take the *.scss URL and return the CSS output.
    pass






import os
import json

example_settings = {   
    # Assetman needs to be able to point at assets to be served by us and by a
    # CDN. We have to serve our own mobile assets in production because of
    # restrictions in offline app caching.

    # Assetman needs to know how to point at assets in three different places:
    # the "normal" static URL (during dev), a "cdn proxy" URL on our own hosts
    # (for, e.g., mobile site or non-CDN fallback), and a set of cdn hosts.
    "static_url_prefix": "/s/",
    "local_cdn_url_prefix": "/cdn/",
    "cdn_url_prefix": [
        '//1.example.net/',
        '//2.example.net/',
        '//3.example.net/',
    ],
    # If we need to fall back to serving all static assets ourselves
    # 'cdn_url_prefix': ['/cdn/'],

    # Where are static files found in the filesystem for this project?
    "static_dir": os.path.abspath(os.path.join(os.path.dirname(__file__), "static")),
    # And where are compiled static assets found?
    "compiled_asset_root": "/data/compiled_assets",
}
default_settings = {
    "closure_compiler":"/bin/closure-compiler.jar",
    "minify_compressor_path": "/bin/minify",
    "sass_compiler": "/bin/sass",
    "lessc_path": "/bin/lessc",
    "java_bin": "java"
}

class Settings(dict):
    """
    A custom dictionary object representing an Assetman configuration.
    Settings may be loaded from a configuration file or initialized through
    the object's creation. If a file is provided, additional arguments will
    serve as overrides to any values found in the file.
    """
    def __init__(self, configuration_file=None, *args, **kwargs):
        self._config_path = configuration_file
        self.update(default_settings)
        self.update(dict(*args, **kwargs))

    @classmethod
    def load(cls, path):
        """
        Loads a settings configuration file, overwriting any existing setting
        values.
        """ 

        with open(path, 'r') as config:
            return Settings(configuration_file=path, **json.loads(config.read()))

    def save(self, path=None):
        """
        Saves the current settings object to its source configuration file. If
        a path is provided, it will save to the path instead.
        """

        usepath = path if path else self._path 

        if not usepath:
            raise Exception("No destination file specified to save.")

        with open(usepath, 'w') as outfile:
            outfile.write(json.dumps(self))





import base64
from collections import defaultdict
import hashlib
import logging
import mimetypes
import subprocess
import functools
import os
import re

import assetman.managers
from assetman.tools import make_absolute_static_path, make_relative_static_path, get_static_pattern, make_output_path

def run_proc(cmd, stdin=None):
    """Runs the given cmd as a subprocess. If the exit code is non-zero, calls
    sys.exit with the exit code (aborting program). If stdin is given, it will
    be piped to the subprocess's stdin.

    The cmd should be a command suitable for passing to subprocess.call (ie, a
    list, usually).
    """
    popen_args = dict(stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if stdin is not None:
        popen_args['stdin'] = subprocess.PIPE
        stdin = stdin.encode()
    proc = subprocess.Popen(cmd, **popen_args)
    out, err = proc.communicate(input=stdin)
    if proc.returncode != 0:
        raise CompileError(cmd, err)
    elif err:
        logging.warning('%s stderr:\n%s', cmd[0], err)
    return out.decode()

class CompileError(Exception):
    """Error encountered while compiling assets."""

class ParseError(Exception):
    """Error encountered while parsing templates."""

class DependencyError(Exception):
    """Invalid or missing dependency."""

class AssetCompiler(object):
    """A base class meant to be mixed in with `assetman.AssetManager`
    subclasses to provide support for compiling asset blocks.
    """

    # What do this compiler's {% assetman.include_* %} expressions look like?
    include_expr = None

    def __init__(self, *args, **kwargs): 
        super(AssetCompiler, self).__init__(*args, **kwargs)

    def required_setting_file(self, key):
        """
        Get the named setting from self.settings and give an informative
        error if it's missing.

        This is helpful because when run under multiprocessing, we get only
        the exception name and message without a useful traceback,
        so it's hard to know how/where to fix the problem.
        http://bugs.python.org/issue13831
        """
        path = self.settings[key]
        assert os.path.exists(path), "missing file %s (settings key %s)" % (path, key)
        return path
        

    def compile(self, **kwargs):
        """Compiles the assets in this Assetman block. Returns compiled source code as a string."""
        logging.info("Compiling %s", self)
        result =  self.do_compile(**kwargs)
        return result

    def do_compile(self, **kwargs):
        raise NotImplementedError

    def needs_compile(self, cached_manifest, current_manifest):
        """Determines whether or not this asset block needs compilation by
        comparing the versions in the given manifests and checking the version
        on disk.
        """
        name_hash = self.get_hash()
        assert name_hash in current_manifest.blocks, self
        content_hash = current_manifest.blocks[name_hash]['version']
        if name_hash in cached_manifest.blocks:
            if cached_manifest.blocks[name_hash]['version'] == content_hash:
                compiled_path = self.get_compiled_path()
                if not os.path.exists(compiled_path):
                    logging.warning('Missing compiled asset %s from %s',
                                 compiled_path, self)
                    return True
                return False
            else:
                logging.warning('Contents of %s changed', self)
        else:
            compiled_path = self.get_compiled_path()
            if not os.path.exists(compiled_path):
                logging.warning('New/unknown hash %s from %s', name_hash, self)
            else:
                logging.info('new hash %s from %s but already exists on file %s', name_hash, self, compiled_path)
                return False
        return True

    def get_current_content_hash(self, manifest):
        """Gets the md5 hash for each of the files in this manager's list of assets."""
        h = hashlib.md5()
        for path in self.get_paths():
            relative_path = make_relative_static_path(self.settings['static_dir'], path)
            assert relative_path in manifest.assets, relative_path
            h.update(manifest.assets[relative_path]['version'].encode())
        return h.hexdigest()

    def get_paths(self):
        """Returns a list of absolute paths to the assets contained in this manager."""
        paths = list(map(functools.partial(make_absolute_static_path, self.settings['static_dir']), self.rel_urls))
        try:
            assert all(map(os.path.isfile, paths))
        except AssertionError:
            missing = [path for path in paths if not os.path.isfile(path)]
            raise DependencyError(self.src_path, ','.join(missing))
        return paths

    def get_compiled_path(self):
        """Creates the output filename for the compiled assets of the given manager."""
        return make_output_path(self.settings['compiled_asset_root'], self.get_compiled_name())


class JSCompiler(AssetCompiler, assetman.managers.JSManager):

    include_expr = 'include_js'

    def do_compile(self, **kwargs):
        """We just hand each of the input paths to the closure compiler and
        let it go to work.
        """
        cmd = [
            self.required_setting_file("java_bin"), '-Xss16m', '-jar', self.required_setting_file("closure_compiler"),
            '--compilation_level', 'SIMPLE_OPTIMIZATIONS',
            '--language_in', 'ECMASCRIPT5',
            ]
        for path in self.get_paths():
            cmd.extend(('--js', path))
        return run_proc(cmd)


class CSSCompiler(AssetCompiler, assetman.managers.CSSManager):

    include_expr = 'include_css'

    def do_compile(self, **kwargs): 
        """Compiles CSS files using the minify compressor. Since the compressor
        will only accept a single input file argument, we have to manually
        concat the CSS files in the batch and pipe them into the compressor.

        https://github.com/tdewolff/minify/tree/master/cmd/minify

        This also allows us to accept a css_input argument, so this function
        can be used by the compile_less function as well.
        """
        css_input = kwargs.get("css_input")
        if css_input is None:
            css_input = '\n'.join(
                open(path).read() for path in self.get_paths())
        if not kwargs.get("skip_inline_images"):
            css_input = self.inline_images(css_input)
        cmd = [
           self.required_setting_file("minify_compressor_path"),
           '--type=css'
        ]
        return run_proc(cmd, stdin=css_input)

    def inline_images(self, css_src):
        """Here we will "inline" any images under a certain size threshold
        into the CSS in the form of "data:" URIs.

        IE 8 can't handle URLs longer than 32KB, so any image whose data URI
        is larger than that is skipped.
        """
        KB = 1024.0
        MAX_FILE_SIZE = 24 * KB # Largest size we consider for inlining
        MAX_DATA_URI_SIZE = 32 * KB # IE8's maximum URL size

        # We only want to replace asset references that show up inside of
        # `url()` rules (this avoids weird constructs like IE-specific filters
        # for transparent PNG support).
        base_pattern = get_static_pattern(self.settings.get('static_url_prefix'))
        pattern = r"""(url\(["']?)%s(["']?\))""" % base_pattern

        # Track duplicate images so that we can warn about them
        seen_assets = defaultdict(int)

        def replacer(match):
            before, url_prefix, rel_path, after = match.groups()
            path = make_absolute_static_path(self.settings['static_dir'], rel_path)
            assert os.path.isfile(path), (path, str(self))
            if os.stat(path).st_size > MAX_FILE_SIZE:
                logging.debug('Not inlining %s (%.2fKB)', path, os.stat(path).st_size / KB)
                return match.group(0)
            else:
                encoded = base64.b64encode(open(path, 'rb').read())
                mime_type, _ = mimetypes.guess_type(path)
                if not mime_type and path.endswith('.otf'):
                    mime_type = 'application/octet-stream'
                if not mime_type and path.endswith('.ttf'):
                    mime_type = 'font/ttf'
                if not mime_type and path.endswith('.eot'):
                    mime_type = 'application/vnd.ms-fontobject'
                if not mime_type and path.endswith('.woff'):
                    mime_type = 'application/x-font-woff'
                if not mime_type and path.endswith('.json'):
                    mime_type = 'application/json'
                if not mime_type and path.endswith('.svg'):
                    mime_type = 'image/svg+xml'
                data_uri = 'data:%s;base64,%s' % (mime_type, encoded)
                if len(data_uri) >= MAX_DATA_URI_SIZE:
                    logging.debug('Not inlining %s (%.2fKB encoded)', path, len(data_uri) / KB)
                    return match.group(0)
                seen_assets['%s%s' % (url_prefix, rel_path)] += 1
                return ''.join([before, data_uri, after])

        result = re.sub(pattern, replacer, css_src)

        for url, count in seen_assets.items():
            if count > 1:
                logging.warning('inline asset duplicated %dx: %s', count, url)

        return result


class LessCompiler(CSSCompiler, assetman.managers.LessManager):

    include_expr = 'include_less'

    def do_compile(self, **kwargs):
        """Compiling less files is an ugly 2-step process, because the lessc
        compiler sucks.

        First, we have to run each of the given paths through lessc
        separately, capturing and concatenating the output. Then, we send all
        of the compiled CSS to the YUI compressor.
        """
        # First we "compile" the less files into CSS
        lessc = self.required_setting_file("lessc_path")
        outputs = [run_proc([lessc, path]) for path in self.get_paths()]
        return super(LessCompiler, self).do_compile(css_input='\n'.join(outputs))


class SassCompiler(CSSCompiler, assetman.managers.SassManager):

    include_expr = 'include_sass'

    def do_compile(self, **kwargs):
        cmd = [
            self.required_setting_file("sass_compiler_path"),
            '--compass', '--trace', '--no-cache', '--stop-on-error', '-l'
        ] + self.rel_urls
        output = run_proc(cmd)
        return super(SassCompiler, self).do_compile(css_input=output)






import os
import re
import binascii
import itertools
import logging

# What to calls to assetman look like in {% apply %} blocks?
include_expr_matcher = re.compile(r'^assetman\.(include_\w+)').match
def get_shard_from_list(settings_list, shard_id):
    assert isinstance(settings_list, (list, tuple)), "must be a list not %r" % settings_list
    shard_id = _crc(shard_id)
    bucket = shard_id % len(settings_list)
    return settings_list[bucket]

def _crc(key):
    """crc32 hash a string"""
    return binascii.crc32(_utf8(key)) & 0xffffffff

def _utf8(s):
    """encode a string as utf-8, returning bytes"""
    if isinstance(s, str):
        return s.encode("utf-8")
    if isinstance(s, bytes):
        try:
            s.decode("utf-8")
        except UnicodeDecodeError:
            raise AssertionError("Invalid encoding. _utf8 expected a str or utf-8 encoded bytes")
    assert isinstance(s, bytes), "_utf8 expected a str or utf-8 encoded bytes, not %r" % type(s)
    return s

def _unicode(value):
    """decode utf-8 bytes, returning string"""
    if isinstance(value, bytes):
        try:
            return value.decode("utf-8")
        except UnicodeDecodeError:
            raise AssertionError("Invalid encoding. _unicode expected a str or utf-8 encoded bytes")
    assert isinstance(value, str), "_unicode expected a str or utf-8 encoded bytes, not %r" % type(value)
    return value

def iter_template_paths(template_dirs, template_ext):
    """Walks each directory in the given list of template directories,
    yielding the path to each template found.
    """
    # We only try to parse files that match this pattern as Tornado templates
    template_file_matcher = re.compile(r'\.' + re.escape(template_ext) + '$').search

    for template_dir in template_dirs:
        for root, dirs, files in os.walk(template_dir):
            for f in filter(template_file_matcher, files):
                yield os.path.join(root, f)

# Shortcuts for creating paths relative to some other path
def make_absolute_static_path(static_dir, p):
    if os.path.exists(p):
        return p
    return os.path.normpath(os.path.abspath(os.path.join(static_dir, p)))

def make_relative_static_path(static_dir, p):
    base = os.path.normpath(os.path.abspath(static_dir))
    if p.startswith(base):
        return p[len(base)+1:]
    return p

def make_output_path(compiled_asset_root, p):
    return os.path.join(compiled_asset_root, p)

def get_static_pattern(static_url_prefix):
    """Builds a regular expression we can use to find static asset references
    that start with the given static URL prefix.

    Used for finding static dependencies inside JS and CSS files and for
    rewriting static references in compiled assets.
    """
    return r'(%s)(.*?\.\w+)' % re.escape(static_url_prefix)

def get_parser(template_path, template_type, settings):
    """ Factory method to return appropriate parser class """
    #TODO: dynamic import / return based on settings / config
    #avoids circular dep
    assert template_type in ["tornado_template"]
    assert isinstance(template_path, str)
    logging.info('parsing for %s %s', template_type, template_path)
    if settings.get('verbose'):
        print(template_type, template_path)
    from assetman.parsers.tornado_parser import TornadoParser
    return TornadoParser(template_path, settings)



#!/bin/python



import os
import re
import logging
import multiprocessing
import hashlib

from optparse import OptionParser

from assetman.manifest import Manifest 
from assetman.tools import iter_template_paths, get_static_pattern, make_relative_static_path, make_absolute_static_path, get_parser
from assetman.compilers import DependencyError, ParseError, CompileError
from assetman.S3UploadThread import upload_assets_to_s3

from assetman.settings import Settings

class NeedsCompilation(Exception):
    pass

parser = OptionParser(description='Compiles assets for AssetMan')

parser.add_option(
    '--tornado_template_dirs', type="string", action='append', 
    help='Directory to crawl looking for static assets to compile.')

parser.add_option(
    '--template-ext', type="string", default="html",
    help='File extension of compilable templates.')

parser.add_option(
    '--static-dir', type="string", default='static',
    help='Directory where static assets are located in this project.')

parser.add_option(
    '--output-dir', type="string", default='assets',
    help='Directory where compiled assets are to be placed.')

parser.add_option(
    '--static-path-prefix', type="string", default="",
    help="Directory prefix to access static assets (matching url prefix)")

parser.add_option(
    '--static-url-path', type="string", default="/",
    help="Static asset base url path")

parser.add_option(
    '--compiled-manifest-path', type="string", default="/",
    help="Location to read/write the compiled asset manifest")

parser.add_option(
    '-t', '--test-needs-compile', action="store_true",
    help='Check whether a compile is needed. Exits 1 if so.')

parser.add_option(
    '-f', '--force-recompile', action="store_true",
    help='Force a recompile of everything.')

parser.add_option(
    '-i', '--skip-inline-images', action="store_true",
    help='Do not sub data URIs for small images in CSS.')

parser.add_option(
    '--skip-s3-upload', action="store_true",
    help='Skip uploading anything to s3')

parser.add_option(
    '--aws_username', type="string", 
    help="AWS username, required for uploading to s3 (no upload if ommited)")

parser.add_option(
    '--aws_access_key', type="string",
    help="AWS access key, required for uploading to s3")

parser.add_option(
    '--aws_secret_key', type="string",
    help="AWS secret key, required for uplaoding to s3")

parser.add_option(
    '--s3_assets_bucket', type="string",
    help="AWS s3 bucket to store assets, required for uploading to s3")


# Static calls are like {{ assetman.static_url('path.jpg') }} or include
# an extra arg: {{ assteman.static_url('path.jpg', local=True) }}
static_url_call_finder = re.compile(r'assetman\.static_url\((.*?)(,.*?)?\)').finditer

##############################################################################
# Multiprocessing workers
##############################################################################
class ParserWorker(object):
    def __init__(self, settings):
        self.settings = settings

    def __call__(self, template_info):
        """Takes a template path and returns a list of AssetCompiler instances
        extracted from that template. Helper function to be called by each process
        in the process pool created by find_assetman_compilers, above.
        """
        assert isinstance(template_info, (list, tuple))
        template_path, template_type = template_info
        template = get_parser(template_path, template_type, self.settings)
        return list(template.get_compilers())

class CompileWorker(object):
    """Takes an AssetCompiler and, based on the manifest, compiles the assets,
    writing the results to disk. Used as a helper function when compiling
    assets in parallel.
    """
    def __init__(self, skip_inline_images, manifest):
        self.manifest = manifest
        self.skip_inline_images = skip_inline_images

    def __call__(self, compiler):
        with open(compiler.get_compiled_path(), 'w') as outfile:
            outfile.write(compiler.compile(skip_inline_images=self.skip_inline_images))


##############################################################################
# Compiler support functions
##############################################################################
def get_file_hash(path, block_size=8192):
    """Calculates the content hash for the file at the given path."""
    md5 = hashlib.md5()
    with open(path, 'rb') as f:
        while True:
            block = f.read(block_size)
            if not block:
                break
            md5.update(block)
    return md5.hexdigest()

##############################################################################
# Build and compare dependency manifests
##############################################################################
def static_finder(s, static_url_prefix):
    pattern = get_static_pattern(static_url_prefix)
    return re.compile(pattern).finditer(s)

def import_finder(s):
    return re.compile(r"""@import (url\()?(["'])(.*?)(\2)""").finditer(s)

def empty_asset_entry():
    """This is the form of each 'assets' entry in the manifest as we're
    building it.
    """
    return {
        'version': None,
        'versioned_path': None,
        'deps': set()
    }

def build_compilers(path_infos, settings):
    """Parse each template and return a list of AssetCompiler instances for
    any assetman.include_* blocks in each template.
    """
    # pool.map will not ordinarily handle KeyboardInterrupts cleanly,
    # but if you give them a timeout they will. More info:
    # http://bugs.python.org/issue8296
    # http://stackoverflow.com/a/1408476/151221

    parser_worker = ParserWorker(settings)
    # a sync version for easier debugging (to see exceptions)
    compiler_lists = [parser_worker(x) for x in path_infos]
    output = [item for sublist in compiler_lists for item in sublist]
    return output
    # pool = multiprocessing.Pool()
    # return [x for xs in pool.map_async(parser_worker, path_infos).get(1e100) for x in xs]


def iter_template_deps(static_dir, src_path, static_url_prefix):
    """Yields static resources included as {{assetman.static_url()}} calls
    in the given source path, which should be a Tornado template.

    TODO: need one of these for every supported template language?
    """
    src = open(src_path).read()
    for match in static_url_call_finder(src):
        arg = match.group(1)
        quotes = '\'"'
        if arg[0] not in quotes or arg[-1] not in quotes:
            msg = 'Vars not allowed in static_url calls: %s' % match.group(0)
            raise ParseError(src_path, msg)
        else:
            dep_path = make_absolute_static_path(static_dir, arg.strip(quotes))
            if os.path.isfile(dep_path):
                yield dep_path
            else:
                logging.warning('Missing dep %s (src: %s)', dep_path, src_path)

###############################################################################

def version_dependency(path, manifest):
    """A dependency's version is calculated using this recursive formula:

        version = md5(md5(path_contents) + version(deps))

    So, the version of a path is based on the hash of its own file contents as
    well as those of each of its dependencies.
    """
    assert path in manifest.assets, path
    if manifest.assets[path]['version']:
        return manifest.assets[path]['version']
    h = hashlib.md5()
    h.update(get_file_hash(make_absolute_static_path(manifest.settings['static_dir'], path)).encode())
    for dep_path in manifest.assets[path]['deps']:
        h.update(version_dependency(dep_path, manifest).encode())
    version = h.hexdigest()
    _, ext = os.path.splitext(path)
    manifest.assets[path]['version'] = version
    manifest.assets[path]['versioned_path'] = version + ext
    return manifest.assets[path]['version']


def iter_deps(static_dir, src_path, static_url_prefix):
    """Yields first-level dependencies from the given source path."""
    assert os.path.isfile(src_path), src_path
    dep_iter = {
        '.js': iter_js_deps,
        '.css': iter_css_deps,
        '.less': iter_css_deps,
        '.scss': iter_scss_deps,
        '.html': iter_template_deps,
        }.get(os.path.splitext(src_path)[1])
    if dep_iter:
        for path in dep_iter(static_dir, src_path, static_url_prefix):
            yield path

def iter_css_deps(static_dir, src_path, static_url_prefix):
    """Yields first-level dependencies from the given source path, which
    should be a CSS, Less, or Sass file. Dependencies will either be more
    CSS/Less/Sass files or static image resources.
    """
    assert os.path.isfile(src_path), src_path
    root = os.path.dirname(src_path)
    src = open(src_path).read()

    # First look for CSS/Less imports and recursively descend into them
    for match in import_finder(src):
        path = match.group(3)
        if src_path.endswith('.less') and os.path.splitext(path)[1] == '':
            path = path + '.less'
        # normpath will take care of '../' path components
        new_root = os.path.normpath(os.path.join(root, os.path.dirname(path)))
        full_path = os.path.join(new_root, os.path.basename(path))
        assert os.path.isdir(new_root), new_root
        assert os.path.isfile(full_path), full_path
        yield full_path

    # Then look for static assets (images, basically)
    for dep in iter_static_deps(static_dir, src_path, static_url_prefix):
        yield dep

def iter_scss_deps(src_path):
    """Yields first-level dependencies from the given source path, which
    should be a Sass file. Dependencies will either be more
    Sass files or static image resources.
    """
    assert os.path.isfile(src_path), src_path
    root = os.path.dirname(src_path)
    src = open(src_path).read()

    # First look for CSS/Less imports and recursively descend into them
    for match in import_finder(src):
        path = match.group(3)
        if path.startswith('compass/'):
            continue
        # normpath will take care of '../' path components
        new_root = os.path.normpath(os.path.join(root, os.path.dirname(path)))
        full_path = os.path.join(new_root, os.path.basename(path))
        assert os.path.isdir(new_root), new_root
        assert os.path.isfile(full_path), full_path
        yield full_path

    # Then look for static assets (images, basically)
    for dep in iter_static_deps(src_path):
        yield dep


def iter_js_deps(static_dir, src_path, static_url_prefix):
    """Yields first-level dependencies from the given source path, which
    should be a JavaScript file. Dependencies will be static image resources.
    """
    return iter_static_deps(static_dir, src_path, static_url_prefix)

def iter_static_deps(static_dir, src_path, static_url_prefix):
    """Yields static resources (ie, image files) from the given source path.
    """
    assert os.path.isfile(src_path), src_path
    for match in static_finder(open(src_path).read(), static_url_prefix):
        dep_path = make_absolute_static_path(static_dir, match.group(2))
        if os.path.isfile(dep_path):
            yield dep_path
        else:
            logging.warning('Missing dep %s (src: %s)', dep_path, src_path)


def _build_manifest_helper(static_dir, src_paths, static_url_prefix, manifest):
    assert isinstance(src_paths, (list, tuple))
    for src_path in src_paths:
        # Make sure every source path at least has the skeleton entry
        rel_src_path = make_relative_static_path(static_dir, src_path)
        logging.info('_build_manifest_helper %s (current %s)', src_path, manifest.assets.get(rel_src_path))
        manifest.assets.setdefault(rel_src_path, empty_asset_entry())
        for dep_path in iter_deps(static_dir, src_path, static_url_prefix):
            logging.info('%s > dependency %s', src_path, dep_path)
            rel_path = make_relative_static_path(static_dir, dep_path)
            manifest.assets[rel_src_path]['deps'].add(rel_path)
            _build_manifest_helper(static_dir, [dep_path], static_url_prefix, manifest)


def build_manifest(tornado_paths, settings):
    """Recursively builds the dependency manifest for the given list of source
    paths.
    """
    assert isinstance(tornado_paths, (list, tuple))

    paths = list(set(tornado_paths))
    # First, parse each template to build a list of AssetCompiler instances
    path_infos = [(x, 'tornado_template') for x in tornado_paths]
    compilers = build_compilers(path_infos, settings)

    # Add each AssetCompiler's paths to our set of paths to search for deps
    paths = set(paths)
    for compiler in compilers:
        new_paths = compiler.get_paths()
        if settings.get('verbose'):
            print(compiler, new_paths)
        paths.update(new_paths)
    paths = list(paths)

    # Start building the new manifest
    manifest = Manifest(settings)
    _build_manifest_helper(settings['static_dir'], paths, settings['static_url_prefix'], manifest)
    assert all(make_relative_static_path(settings['static_dir'], path) in manifest.assets for path in paths)

    # Next, calculate the version hash for each entry in the manifest
    for src_path in manifest.assets:
        version_dependency(src_path, manifest)

    # Normalize and validate the manifest
    manifest.normalize()

    # Update the 'blocks' section of the manifest for each asset block
    for compiler in compilers:
        name_hash = compiler.get_hash()
        content_hash = compiler.get_current_content_hash(manifest)
        manifest.blocks[name_hash] = {
            'version': content_hash,
            'versioned_path': content_hash + '.' + compiler.get_ext(),
        }
        
    return manifest, compilers

def _create_settings(options):
    return Settings(compiled_asset_root=options.output_dir,
                    static_dir=options.static_dir,
                    static_url_prefix=options.static_url_path,
                    tornado_template_dirs=options.tornado_template_dirs,
                    template_extension=options.template_ext,
                    test_needs_compile=options.test_needs_compile,
                    skip_s3_upload=options.skip_s3_upload,
                    force_s3_upload=False,
                    force_recompile=options.force_recompile,
                    skip_inline_images=options.skip_inline_images,
                    aws_username=options.aws_username,
                    aws_access_key=options.aws_access_key,
                    aws_secret_key=options.aws_secret_key,
                    verbose=False,
                    s3_assets_bucket=options.s3_assets_bucket)

def run(settings):
    if not re.match(r'^/.*?/$', settings.get('static_url_prefix')):
        raise Exception('static_url_prefix setting must begin and end with a slash')

    if not os.path.isdir(settings['compiled_asset_root']) and not settings['test_needs_compile']:
        logging.info('Creating output directory: %s', settings['compiled_asset_root'])
        os.makedirs(settings['compiled_asset_root'])

    for d in settings['tornado_template_dirs']:
        if not os.path.isdir(d):
            raise Exception('Template directory not found: %r', d)

    if not os.path.isdir(settings['static_dir']):
        raise Exception('Static directory not found: %r', settings['static_dir'])

    # Find all the templates we need to parse
    tornado_paths = list(iter_template_paths(settings['tornado_template_dirs'], settings['template_extension']))

    if not tornado_paths:
        logging.warning("No templates found")

    # Load the current manifest and generate a new one
    cached_manifest = Manifest(settings).load()
    try:
        current_manifest, compilers = build_manifest(tornado_paths, settings)
    except ParseError as e:
        src_path, msg = e.args
        logging.error('Error parsing template %s', src_path)
        logging.error(msg)
        raise Exception
    except DependencyError as e:
        src_path, missing_deps = e.args
        logging.error('Dependency error in source %s!', src_path)
        logging.error('Missing paths: %s', missing_deps)
        raise Exception("dependency error in %s. missing %s" % (src_path, missing_deps))

    # Remove duplicates from our list of compilers. This de-duplication must
    # happen after the current manifest is built, because each non-unique
    # compiler's source path figures into the dependency tracking. But we only
    # need to actually compile each block once.
    logging.debug('Found %d assetman block compilers', len(compilers))
    compilers = list(dict((c.get_hash(), c) for c in compilers).values())
    logging.debug('%d unique assetman block compilers', len(compilers))

    # update the manifest on each our compilers to reflect the new manifest,
    # which is needed to know the output path for each compiler.
    for compiler in compilers:
        compiler.manifest = current_manifest

    # Figure out which asset blocks need to be (re)compiled, if any.
    def needs_compile(compiler):
        return compiler.needs_compile(cached_manifest, current_manifest)

    if settings['force_recompile']:
        to_compile = compilers
    else:
        to_compile = list(filter(needs_compile, compilers))

    if to_compile or cached_manifest.needs_recompile(current_manifest):
        # If we're only testing whether a compile is needed, we're done
        if settings['test_needs_compile']:
            raise NeedsCompilation()

        pool = multiprocessing.Pool()
        try:
            # See note above about bug in pool.map w/r/t KeyboardInterrupt.
            _compile_worker = CompileWorker(settings.get('skip_inline_images', False), current_manifest)
            pool.map_async(_compile_worker, to_compile).get(1e9) # previously set to 1e100 which caused overflow of C _PyTime_t_
        except CompileError as e:
            cmd, msg = e.args
            logging.error('Compile error!')
            logging.error('Command: %s', ' '.join(cmd))
            logging.error('Error:   %s', msg)
            raise Exception('Compilation Failed')

        #TODO: refactor to some chain of command for plugins
        if settings['aws_username']:
            upload_assets_to_s3(current_manifest, settings, skip_s3_upload=settings['skip_s3_upload'])
        

        if settings.get('merge_manifest_updates', True):
            cached_manifest.union(current_manifest)
        else:
            cached_manifest = current_manifest
        cached_manifest.write(settings=settings)
        return cached_manifest
    return current_manifest

if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    options, args = parser.parse_args()
    settings = _create_settings(options) 
    run(settings)





import os
import json
import logging

from assetman.settings import Settings

class Manifest(object):
    """
    Represents the asset manifest. 
    TODO: Currently this only handles the utility reading and writing to file functions. 
    Long-term goal is to create a complete model representation of the manifest.
    """

    _manifest = None

    def __init__(self, settings=None):
        self.settings = settings or Settings()
        self._manifest = self.make_empty_manifest()

    def data(self):
        return self._manifest

    @property
    def assets(self):
        return self._manifest['assets']
    @property
    def blocks(self):
        return self._manifest['blocks']
    
    def __str__(self):
        return '<Manifest %s assets:%s blocks:%s>' % (self.get_path(), self.assets, self.blocks)

    @classmethod
    def wrap(cls, data):
        assert isinstance(data, dict)
        assert 'assets' in data
        assert 'blocks'in data
        m = cls()
        m._manifest = data
        return m

    def get_path(self, compiled_asset_path=None):
        compiled_asset_path = compiled_asset_path or self.settings["compiled_asset_root"]
        return os.path.join(compiled_asset_path, 'manifest.json')

    def load(self, compiled_asset_path=None):
        try:
            filename = self.get_path(compiled_asset_path)
            self._manifest = json.load(open(filename))
            assert isinstance(self.assets, dict)
            assert isinstance(self.blocks, dict)
        except (AssertionError, KeyError, IOError, json.JSONDecodeError) as e:
            logging.warning('error opening manifest file: %s', e)
            self._manifest = self.make_empty_manifest()

        return self

    def write(self, compiled_asset_path=None, settings=None, **kwargs):
        if settings is not None:
            self.settings = settings
        manifest_path = self.get_path(compiled_asset_path)
        logging.info('Writing manifest to %s', manifest_path)
        json.dump(self._manifest, open(manifest_path, 'w'), indent=2, **kwargs)

    def make_empty_manifest(self):
        return {
            'blocks': {},
            'assets': {},
        }

    def normalize(self):
        """Normalizes and sanity-checks the given dependency manifest by first
        ensuring that all deps are expressed as lists instead of sets (as they are
        when the manifest is built) and then by ensuring that every dependency has
        its own entry and version in the top level of the manifest.
        """
        for parent, depspec in self.assets.items():
            depspec['deps'] = list(depspec['deps'])
            for dep in depspec['deps']:
                assert dep in self.assets, (parent, dep)
                assert depspec['version'], (parent, dep)
        for name_hash, depspec in self.blocks.items():
            assert depspec['version'], name_hash

    def union(self, newer_manifest):
        # add an age entry to the old side so we can know how many merge
        # generations ago an entry was created
        def age(entry):
            entry['age'] = entry.get('age', 0) + 1
        list(map(age, list(self.blocks.values())))
        list(map(age, list(self.assets.values())))
        self.blocks.update(newer_manifest.blocks)
        self.assets.update(newer_manifest.assets)
    
    def needs_recompile(self, newer_manifest):
        # Figure out if any static assets referenced in the new manifest are
        # missing from the cached manifest.
        def assets_in_sync(asset):
            if asset not in self.assets:
                logging.info('new asset %s (not in current manifest)', asset)
                return False
            if self.assets[asset]['version'] != newer_manifest.assets[asset]['version']:
                logging.warning('Static asset %s version mismatch', asset)
                return False
            return True
        assets_out_of_sync = not all(map(assets_in_sync, newer_manifest.assets))
        if assets_out_of_sync:
            logging.warning('Static assets out of sync')
        return assets_out_of_sync




#!/bin/python

import re
import os
import os.path
import sys
import threading
import datetime
import queue as Queue
import mimetypes
import logging
import boto3
from assetman.tools import make_output_path, make_absolute_static_path, make_relative_static_path, get_static_pattern, get_shard_from_list, _unicode

class S3UploadThread(threading.Thread):
    """Thread that knows how to read asset file names from a queue and upload
    them to S3. Any exceptions encountered will be added to a shared errors
    list.

    Each asset will be uploaded twice:  Once with a special "/cdn/" prefix for
    assets to be served by a proxy on our own domain, and again without the
    prefix for assets to be served via CloudFront.  For the second upload,
    static URLs inside each asset will have to be rewritten *again* to point
    at CloudFront instead of our local CDN proxy.
    """

    def __init__(self, queue, errors, manifest, settings):
        threading.Thread.__init__(self)
        self.client = boto3.client('s3',
            aws_access_key_id=settings.get('aws_access_key'),
            aws_secret_access_key=settings.get('aws_secret_key'))
        self.bucket = boto3.resource('s3',
            aws_access_key_id=settings.get('aws_access_key'),
            aws_secret_access_key=settings.get('aws_secret_key')).Bucket(settings.get('s3_assets_bucket'))
        self.queue = queue
        self.errors = errors
        self.manifest = manifest
        self.settings = settings

    def run(self):
        while True:
            file_name, file_path = self.queue.get()
            try:
                self.start_upload_file(file_name, file_path)
            except Exception as e:
                logging.error('Error uploading %s: %s', file_name, e)
                self.errors.append((sys.exc_info(), self))
            finally:
                self.queue.task_done()

    def start_upload_file(self, file_name, file_path):
        """Starts the procecss of uploading a file to S3. Each file will be
        uploaded twice (once for CDN and once for our local CDN proxy).
        """
        assert isinstance(file_name, str)
        assert isinstance(file_path, str)
        assert os.path.isfile(file_path)

        content_type, content_encoding = mimetypes.guess_type(file_name)
        if not content_type:
            ext = os.path.splitext(file_name)[-1]
            content_type = {
                '.woff': 'application/font-woff',
                '.ttf': 'font/ttf',
                '.otf': 'font/opentype',
                '.eot': 'application/vnd.ms-fontobject',
                '.svg': 'image/svg+xml',
            }.get(ext, 'application/octet-stream')
        headers = {
            'Content-Type': content_type,
            'Cache-Control': self.get_cache_control(),
        }

        with open(file_path, 'rb') as f:
            file_data = f.read()
            # First we will upload the asset for serving via CloudFront CDN,
            # so its S3 key will not have a prefix.
            key = self.bucket.Object(file_name)
            self.upload_file(key, file_data, headers, for_cdn=True)

            # Next we will upload the same file with a prefixed key, to be
            # served by our "local CDN proxy".
            key_prefix = self.settings.get('local_cdn_url_prefix').lstrip('/').rstrip('/')
            key = self.bucket.Object(key_prefix + '/' + file_name)
            self.upload_file(key, file_data, headers, for_cdn=False)

    def exists(self, obj):
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.head_object
        try:
            self.client.head_object(Bucket=obj.bucket_name, Key=obj.key)
        except Exception as e:
            logging.error('got %s when searching for %s', e, obj.key)
            return False
        return True

    def upload_file(self, key, file_data, headers, for_cdn):
        """Uploads the given file_data to the given S3 key. If the file is a
        compiled asset (ie, JS or CSS file), any static URL references it
        contains will be rewritten before upload.

        If use_cdn is True, static URL references will be updated to point to
        our CloudFront CDN domains. Otherwise, they will be updated to point
        to our local CDN proxy.
        """
        if self.settings.get('force_s3_upload') or not self.exists(key):
            # Do we need to do URL replacement?
            if re.search(r'\.(css|js)$', key.key):
                if for_cdn:
                    logging.info('Rewriting URLs => CDN in %s', key.key)
                    replacement_prefix = self.settings.get('cdn_url_prefix')
                else:
                    logging.info('Rewriting URLs => local proxy in %s', key.key)
                    replacement_prefix = self.settings.get('local_cdn_url_prefix')
                file_data = sub_static_version(
                    file_data,
                    self.manifest,
                    replacement_prefix,
                    self.settings['static_dir'],
                    self.settings.get('static_url_prefix'))
            key.put(Body=file_data, CacheControl=headers.get('Cache-Control'), ContentType=headers.get('Content-Type'), ACL="public-read", Expires=self.get_expires())
            logging.info('Uploaded s3://%s/%s', key.bucket_name, key.key)
            logging.debug('Headers: %r', headers)
        else:
            logging.info('Skipping upload of %s; already exists (use force_s3_upload to override)', key.key)

    def get_expires(self):
        # Get a properly formatted date and time, via Tornado's set_header()
        dt = datetime.datetime.utcnow() + datetime.timedelta(days=365*10)
        return dt


    def get_cache_control(self):
        return 'public, max-age=%s' % (86400 * 365 * 10)


def upload_assets_to_s3(manifest, settings, skip_s3_upload=False):
    """Uploads any assets that are in the given manifest and in our compiled
    output dir but missing from our static assets bucket to that bucket on S3.
    """
    # We will gather a set of (file_name, file_path) tuples to be uploaded
    to_upload = set()

    # We know we want to upload each asset block (these correspond to the
    # assetman.include_* blocks in each template)
    for depspec in manifest.blocks.values():
        file_name = depspec['versioned_path']
        file_path = make_output_path(settings['compiled_asset_root'], file_name)
        assert os.path.isfile(file_path), 'Missing compiled asset %s' % file_path
        to_upload.add((file_name, file_path))

    # And we know that we'll want to upload any statically-referenced assets
    # (from assetman.static_url calls or referenced in any compiled assets),
    # but we'll need to filter out other entries in the complete 'assets'
    # block of the manifest.
    should_skip = re.compile(r'\.(scss|less|css|js|html)$', re.I).search
    for rel_path, depspec in manifest.assets.items():
        if should_skip(rel_path):
            continue
        file_path = make_absolute_static_path(settings['static_dir'], rel_path)
        assert os.path.isfile(file_path), 'Missing static asset %s' % file_path
        file_name = depspec['versioned_path']
        to_upload.add((file_name, file_path))

    logging.info('Found %d assets to upload to S3', len(to_upload))
    if skip_s3_upload:
        logging.info('Skipping asset upload to S3 %s', to_upload)
        return to_upload

    # Upload assets to S3 using 5 threads
    queue = Queue.Queue()
    errors = []
    for i in range(5):
        uploader = S3UploadThread(queue, errors, manifest, settings)
        uploader.setDaemon(True)
        uploader.start()
    list(map(queue.put, to_upload))
    queue.join()
    if errors:
        raise Exception(errors)
    return to_upload


def sub_static_version(src, manifest, replacement_prefix, static_dir, static_url_prefix):
    """Adjusts any static URLs in the given source to point to a different
    location.

    Static URLs are determined based on the the 'static_url_prefix' setting.
    They will be updated to point to the given replacement_prefix, which can
    be a string or a list of strings (in which case the actual replacement
    prefix will be chosen by sharding each asset's base name).
    """
    def replacer(match):
        prefix, rel_path = match.groups()
        path = make_relative_static_path(static_dir, rel_path)
        if path in manifest.assets:
            versioned_path = manifest.assets[path]['versioned_path']
            if isinstance(replacement_prefix, (list, tuple)):
                prefix = get_shard_from_list(replacement_prefix, versioned_path)
            else:
                prefix = replacement_prefix
            replacement_link = prefix.rstrip('/') + '/' + versioned_path.lstrip('/')
            logging.info('replacing %s -> %s', path, replacement_link)
            return replacement_link
        logging.warning('Missing path %s in manifest, using %s', path, match.group(0))
        return match.group(0)
    pattern = get_static_pattern(static_url_prefix)
    return re.sub(pattern, replacer, _unicode(src))
